##############################################################################
#
# Copyright (c) 2004 Chris McDonough, Paul Winkler, Jens Vagelpohl and
# Contributors.  All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import os
import time
import tempfile
import thread
import stat
import sys

from ThreadLock import allocate_lock
from ZODB.POSException import ReadConflictError, ConflictError
from Persistence import Persistent
import Globals
import zLOG
from AccessControl import ClassSecurityInfo
import ExtensionClass

from interfaces import IBlob

from uuidgen import new_uuid

ENTRIES = {}
ENTRIES_LOCK = allocate_lock()

class FileBlob(Persistent, ExtensionClass.Base):
    meta_type = "File Blob"

    security = ClassSecurityInfo()

    __allow_access_to_unprotected_subobjects__ = True
    
    __implements__ = IBlob

    def __init__(self):
        self.uid = new_uuid()

    security.declareProtected('ModifyPortalContent', 'open')
    def open(self, mode='rb'):
        return local_file_adapter.open(self, mode)

    security.declareProtected('ModifyPortalContent', 'unlink')
    def unlink(self):
        self.open('u')

Globals.InitializeClass(FileBlob)

def filename_from_blob(blob, datadir):
    uid = blob.uid
    nonblank = filter(None, [datadir, uid[0:2], uid[2:4], uid[4:]])
    path =  os.path.join(*nonblank)
    return path

class FileAdapter:
    def __init__(self, datadir, tempdir=None):
        self.datadir = datadir
        self.tempdir = tempdir

    def open(self, blob, mode):
        """ XXX We make the assumption that the time that the blob's
        database connection was opened is the start time of the
        enclosing transaction.  This is usually true (or at least
        close enough), but it can be wrong when an application does
        its own transaction management instead of using Zope's default
        transaction mgmt.  Transactions themselves do not keep track
        of their start times, or we would use that time value
        instead. """
        path = filename_from_blob(blob, self.datadir)
        if blob._p_jar is not None:
            txn_start = blob._p_jar._opened
        else:
            txn_start = time.time() # XXX what to do when blob has no jar?
        entry = Entry(self, path, mode, txn_start)
        return entry.open()

local_file_adapter = FileAdapter(CLIENT_HOME)

Unlinked = () # marker

class Entry:
    def __init__(self, adapter, path, mode, txn_start):
        self.adapter = adapter
        self.binary = False
        newmode = mode.replace('b', '')
        if newmode != mode:
            mode = newmode
            self.binary = True
        if not mode in ['r', 'w', 'a', 'u']:
            raise IOError, "invalid mode flag(s): %s" % mode
        self.path = path
        self.mode = mode
        self.txn_start = txn_start
        self.temppath = None
        self.file = None
        self.transaction_done = False
        self.op = None

    def open(self):
        self.timestamp = time.time()
        get_transaction().register(self)
        method = getattr(self, '_%s' % self.mode)
        self.file = method()
        return self.file

    def has_changed(self):
        try:
            # The Python os.stat API has a one-second resolution for
            # time values , so we need to take that into account when
            # reporting whether the file has changed or not.  This
            # will cause inappropriate conflict errors, but the
            # alternative is to do file content checks, locking, or
            # write our own mtime-gatherer that takes into account
            # st_mtime_usec (POSIX).
            f_mod_time = os.stat(self.path)[stat.ST_MTIME] + 1
        except OSError: # file doesnt exist
            f_mod_time = 0

        if f_mod_time > self.txn_start:
            return True

    def _r(self):
        mode = 'r'
        if self.binary:
            mode = mode + 'b'
        path = self.path
        ENTRIES_LOCK.acquire()
        try:
            temp = ENTRIES.get(thread.get_ident(), {}).get(self.path)
            if temp is Unlinked:
                raise IOError, "No such file or directory"
            if temp is not None:
                path = temp
        finally:
            ENTRIES_LOCK.release()
        wrapper = FileWrapper(open(path, mode), path, mode)
        if self.has_changed():
            raise ReadConflictError(DummyPersistent(self.path))
        return wrapper
    
    def _w(self):
        mode = 'w'
        if self.binary:
            mode = mode + 'b'
        if self.adapter.tempdir:
            fd, name = tempfile.mkstemp(dir=self.adapter.tempdir)
        else:
            fd, name = tempfile.mkstemp()
        self.temppath = name
        f = os.fdopen(fd, mode, -1)
        wrapper = FileWrapper(f, name, mode)
        if self.has_changed():
            raise ReadConflictError(DummyPersistent(self.path))
        self.op = 'write'
        ENTRIES_LOCK.acquire()
        try:
            d = ENTRIES.setdefault(thread.get_ident(), {})
            d[self.path] = name
        finally:
            ENTRIES_LOCK.release()
        return wrapper

    def _a(self):
        mode = 'a'
        if self.binary:
            mode = mode + 'b'
        if self.adapter.tempdir:
            fd, name = tempfile.mkstemp(dir=self.adapter.tempdir)
        else:
            fd, name = tempfile.mkstemp()
        self.temppath = name
        f = os.fdopen(fd, mode, -1)

        path = self.path
        ENTRIES_LOCK.acquire()
        try:
            temp = ENTRIES.get(thread.get_ident(), {}).get(self.path)
            if temp is not None and temp is not Unlinked:
                path = temp
        finally:
            ENTRIES_LOCK.release()
        if os.path.exists(path):
            rmode = 'r'
            if self.binary:
                rmode = rmode + 'b'
            actual = open(path, rmode)
            while 1:
                data = actual.read(1 << 16)
                if not data:
                    break
                f.write(data)
            f.flush()
        if self.has_changed():
            raise ReadConflictError(DummyPersistent(self.path))
        self.op = 'append'
        ENTRIES_LOCK.acquire()
        try:
            d = ENTRIES.setdefault(thread.get_ident(), {})
            d[self.path] = name
        finally:
            ENTRIES_LOCK.release()
        return f

    def _u(self):
        """ Delete the cache file from disk """
        self.op = 'unlink'
        ENTRIES_LOCK.acquire()
        try:
            d = ENTRIES.setdefault(thread.get_ident(), {})
            d[self.path] = Unlinked
        finally:
            ENTRIES_LOCK.release()

    #######################################################
    # ZODB Transaction hooks
    #######################################################

    def commit(self, reallyme, t):
        """ Called for every (sub-)transaction commit """
        pass

    def tpc_begin(self, transaction):
        """ Called at the beginning of a transaction """
        pass

    def tpc_abort(self, transaction):
        """ Called on abort - but not always :( """
        pass

    def abort(self, reallyme, t):
        """ Called if the transaction has been aborted """
        if self.temppath:
            # If this was a write operation we need to remove the 
            # tempfile that's still hanging around.
            try:
                os.unlink(self.temppath)
            except (OSError, IOError):
                pass
        if self.file is not None:
            self.file.close()
        try:
            del ENTRIES[thread.get_ident()]
        except KeyError:
            pass

    def tpc_vote(self, transaction):
        """ Only called for real transactions, not subtransactions """
        self.transaction_done = True
        if self.has_changed():
            raise ConflictError(message="Blob conflict for %s" % self.path,
                                object=DummyPersistent(self.path))

    def tpc_finish(self, transaction):
        """ Called at the end of a successful transaction """
        if self.file is not None:
            self.file.close()
        if self.transaction_done and self.op == 'unlink':
            try:
                if os.path.exists(self.path): # XXX race?
                    os.unlink(self.path)
            except IOError:
                zLOG.LOG( 'blob'
                          , zLOG.ERROR
                          , 'IOError removing file "%s"' % self.path
                          , error=sys.exc_info()
                          )
        elif self.transaction_done and self.op in ['write', 'append']:
            # Now rename the tempfile to reflect the desired name.
            # This may fail if they are not on the same filesystem.
            try:
                parent = os.path.realpath(os.path.dirname(self.path))
                if not os.path.isdir(parent) and os.path.exists(parent):
                    zLOG.LOG(
                        'blob',
                        zLOG.ERROR,
                        'IOError renaming file "%s" to "%s"' % (
                        self.file.name, self.path),
                        'The destination path exists'
                        )
                    return

                if not os.path.isdir(parent):
                    try:
                        os.makedirs(parent)
                    except IOError:
                        zLOG.LOG(
                            'blob',
                            zLOG.ERROR,
                            'IOError making dirs for file "%s"'% self.path,
                            error=sys.exc_info()
                            )
                        return

                # XXX Windows will fail here if target path exists, shrug
                os.rename(self.temppath, self.path)
            except IOError:
                 zLOG.LOG( 'blob'
                         , zLOG.ERROR
                         , 'IOError renaming file "%s" to "%s"' % (
                                    self.temppath, self.path)
                         , error=sys.exc_info()
                         )
        try:
            del ENTRIES[thread.get_ident()]
        except KeyError:
            pass

    def sortKey(self, *ignored):
        """ The sortKey method is used for recent ZODB compatibility which
            needs to have a known commit order for lock acquisition.  We use
            our timestamp, which will ensure that the order in which file
            cache entries are created will be the way the operations they
            imply are played back.
        """
        return self.timestamp

class DummyPersistent:
    def __init__(self, path):
        self._p_jar = path
        self._p_oid = None

class FileWrapper(ExtensionClass.Base):
    """File wrapper"""
    __allow_access_to_unprotected_subobjects__ = True
    read__roles__ = None
    readlines__roles__ = None
    readline__roles__ = None
    seek__roles__ = None
    tell__roles__ = None
    fileno__roles__ = None
    flush__roles__ = None
    next__roles__ = None
    write__roles__ = None
    writelines__roles__ = None
    truncate__roles__ = None
    close__roles__ = None

    def __init__(self, file, name, mode):
        self.file = file
        self.name = name
        self.mode = mode
        self.close_called = False

    def __guarded_getattr__(self, name):
        return getattr(self, name)

    def __getattr__(self, name):
        file = self.__dict__['file']
        a = getattr(file, name)
        return a

    def close(self):
        if not self.close_called:
            self.close_called = True
            self.file.close()

Globals.InitializeClass(FileWrapper)

