##############################################################################
#
# Copyright (c) 2004 Chris McDonough, Paul Winkler, Jens Vagelpohl and
# Contributors.  All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest
import os
import time
from Interface.Verify import verifyClass
from ZODB.POSException import ConflictError, ReadConflictError

from Products.blob.interfaces import IBlob
from Products.blob.file import FileBlob
from Products.blob.file import local_file_adapter
from Products.blob.file import filename_from_blob

DATA_DIR = '/tmp/blobtest'
TEMP_DIR = '/tmp/blobtest/tmp'

class DummyJar:
    def __init__(self):
        self._opened = time.time()

class BlobTests(unittest.TestCase):
    def setUp(self):
        try:
            os.makedirs(TEMP_DIR)
            os.makedirs(DATA_DIR)
        except OSError:
            # already exists
            pass
        self.old_datadir = local_file_adapter.datadir
        self.old_tempdir = local_file_adapter.tempdir
        local_file_adapter.datadir = DATA_DIR
        local_file_adapter.tempdir = TEMP_DIR
        
    def tearDown(self):
        local_file_adapter.datadir = self.old_datadir
        local_file_adapter.tempdir = self.old_tempdir
        for root, dirs, files in os.walk(DATA_DIR, topdown=False):
            for name in files:
                os.unlink(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))

    def testBlobImplements(self):
        blob = FileBlob()
        self.failUnless(verifyClass(IBlob, FileBlob))

    def testBlobCreate(self):
        get_transaction().begin()
        blob = FileBlob()
        file = blob.open('wb')
        file.write('foo')
        file.close()
        self.assertEqual(open(file.name).read(), 'foo')
        get_transaction().abort()

    def testBlobReadNonexistent(self):
        blob = FileBlob()
        self.assertRaises(IOError, blob.open, 'r')

    def testBlobChangeAfterCommit(self):
        get_transaction().begin()
        blob = FileBlob()
        self.assertRaises(IOError, open, filename_from_blob(blob, DATA_DIR))
        file = blob.open('wb')
        file.write('foo')
        file.close()
        get_transaction().commit()
        self.assertEqual(open(filename_from_blob(blob, DATA_DIR)).read(),'foo')
        self.failIf(os.path.exists(file.name)) # tempfile should be removed

    def testBlobReadAfterChange(self):
        get_transaction().begin()
        blob = FileBlob()
        self.assertRaises(IOError, open, filename_from_blob(blob, DATA_DIR))
        file = blob.open('wb')
        file.write('foo')
        file.close()
        get_transaction().commit()
        # XXX ensure we sleep long enough to not generate a conflict; this
        # should change in a more robust implementation
        time.sleep(1)
        get_transaction().begin()
        file = blob.open('r')
        self.assertEqual(file.read(), 'foo')

    def testWriteConflict(self):
        blob = FileBlob()
        blob._p_jar = DummyJar()
        get_transaction().begin()
        file = blob.open('wb')
        file.write('foo')
        filename = filename_from_blob(blob, DATA_DIR)
        path = os.path.join(*os.path.split(filename)[:-1])
        os.makedirs(path)
        touch = open(filename, 'wb')
        touch.write('overwritten!')
        touch.close()
        self.assertRaises(ConflictError, get_transaction().commit)

    def testReadConflict(self):
        blob = FileBlob()
        blob._p_jar = DummyJar()
        get_transaction().begin()
        filename = filename_from_blob(blob, DATA_DIR)
        path = os.path.join(*os.path.split(filename)[:-1])
        os.makedirs(path)
        touch = open(filename, 'wb')
        touch.write('overwritten!')
        touch.close()
        self.assertRaises(ReadConflictError, blob.open, 'wb')

    def testAppend(self):
        get_transaction().begin()
        blob = FileBlob()
        file = blob.open('wb')
        file.write('foo')
        file.close()
        get_transaction().commit()
        # XXX ensure we sleep long enough to not generate a conflict; this
        # should change in a more robust implementation
        time.sleep(1)
        get_transaction().begin()
        file = blob.open('a')
        file.write('flea')
        self.assertEqual(open(filename_from_blob(blob, DATA_DIR)).read(),'foo')
        get_transaction().commit()
        self.assertEqual(open(filename_from_blob(blob, DATA_DIR)).read(),
                         'fooflea')
        

    def testUnlink(self):
        get_transaction().begin()
        blob = FileBlob()
        file = blob.open('wb')
        file.write('foo')
        file.close()
        get_transaction().commit()
        # XXX ensure we sleep long enough to not generate a conflict; this
        # should change in a more robust implementation
        time.sleep(1)
        get_transaction().begin()
        self.assertEqual(open(filename_from_blob(blob, DATA_DIR)).read(),'foo')
        blob.unlink()
        self.assertEqual(open(filename_from_blob(blob, DATA_DIR)).read(),'foo')
        get_transaction().commit()
        self.failIf(os.path.exists(filename_from_blob(blob, DATA_DIR)))

    def testChangesRecognizedInTxn(self):
        get_transaction().begin()
        blob = FileBlob()
        file = blob.open('wb')
        file.write('foo')
        file.close()
        file = blob.open('rb')
        self.assertEqual(file.read(), 'foo')
        file.close()
        file = blob.open('ab')
        file.write('fleabar')
        file.close()
        file = blob.open('rb')
        self.assertEqual(file.read(), 'foofleabar')
        get_transaction().abort()

def test_suite():
    suite = unittest.TestSuite((
        unittest.makeSuite(BlobTests),
        ))
    
    return suite

def run():
    unittest.TextTestRunner().run(test_suite())

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')

